package nl.cwi.sen1.tunit;

import java.lang.management.ManagementFactory;
import java.lang.management.MemoryMXBean;
import java.lang.management.ThreadInfo;
import java.lang.management.ThreadMXBean;
import java.net.InetAddress;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import toolbus.ToolBus;
import toolbus.adapter.AbstractTool;
import toolbus.adapter.ToolBridge;
import toolbus.logging.ILogger;
import toolbus.logging.IToolBusLoggerConstants;
import toolbus.logging.LoggerFactory;
import aterm.ATerm;
import aterm.ATermAppl;
import aterm.ATermList;
import aterm.pure.PureFactory;

public class TestToolBridge extends ToolBridge{
	public final static long DEFAULTTIMEOUT = 10000;

	private final Object completionLock = new Object();
	private final WaitLock expectLock = new WaitLock();
	private final WaitLock actionLock = new WaitLock();
	
	private final TUnitTestCase testCase;
	
	private final PureFactory termFactory;
	private final AbstractTool tool;

	private final Map<String, WaitLock> expectedAckEvents;
	private final Map<ATerm, ATerm> expectedEvals;
	private final Set<ATerm> expectedDos;
	
	private final List<ATerm> actionQueue;
	private volatile boolean expecting;
	
	private final boolean verbose;
	
	public TestToolBridge(TUnitTestCase testCase, PureFactory termFactory, AbstractTool tool, String toolName, int toolID, InetAddress host, int port, boolean verbose){
		super(termFactory, toolName, toolID, host, port);
		this.testCase = testCase;
		this.termFactory = termFactory;
		this.tool = tool;
		
		expectedAckEvents = Collections.synchronizedMap(new HashMap<String, WaitLock>());
		expectedEvals = Collections.synchronizedMap(new HashMap<ATerm, ATerm>());
		expectedDos = Collections.synchronizedSet(new HashSet<ATerm>());
		
		actionQueue = Collections.synchronizedList(new ArrayList<ATerm>());
		expecting = false;
		
		this.verbose = verbose;
	}
	
	public TestToolBridge(TUnitTestCase testCase, PureFactory termFactory, AbstractTool tool, String toolName, int toolID, ClassLoader classLoader, ToolBus toolbus, boolean verbose){
		super(termFactory, toolName, toolID, toolbus);
		
		this.testCase = testCase;
		this.termFactory = termFactory;
		this.tool = tool;
		
		expectedAckEvents = Collections.synchronizedMap(new HashMap<String, WaitLock>());
		expectedEvals = Collections.synchronizedMap(new HashMap<ATerm, ATerm>());
		expectedDos = Collections.synchronizedSet(new HashSet<ATerm>());
		
		actionQueue = Collections.synchronizedList(new ArrayList<ATerm>());
		expecting = false;
		
		this.verbose = verbose;
	}
	
	public void receive(byte operation, ATerm aTerm){
		if(operation == ACKEVENT){
			String eventName = ((ATermAppl)((ATermList) aTerm).getFirst()).getAFun().toString();
			WaitLock eventLock = expectedAckEvents.remove(eventName);
			
			synchronized(completionLock){
				completionLock.notifyAll();
			}
			
			if(eventLock == null) testCase.testFailed("Tool "+getToolName()+": unexpected ack event: "+aTerm);
			
			synchronized(eventLock){
				eventLock.notified = true;
				eventLock.notify();
			}
		}
		
		if(verbose) System.err.println("Tool "+getToolName()+" recieved (" + operation + ") " + aTerm);
		
		super.receive(operation, aTerm);
	}
	
	public boolean checkSignature(ATerm signature){
		return true;
	}
	
	public void doDo(ATerm aTerm){
		synchronized(actionLock){
			if(expecting){
				if(!expectedDos.contains(aTerm)) testCase.testFailed("Tool "+getToolName()+": unexpected do: "+aTerm);
			}else{
				actionQueue.add(aTerm);
				
				long waitToMillis = System.currentTimeMillis() + DEFAULTTIMEOUT;
				while(!actionLock.notified && System.currentTimeMillis() < waitToMillis){
					try{
						long timeToWait = waitToMillis - System.currentTimeMillis();
						if(timeToWait > 0) actionLock.wait(timeToWait);
					}catch(InterruptedException irex){/* Ignore this.*/}
				}
				if(!expectedDos.contains(aTerm)) testCase.testFailed("Tool "+getToolName()+": unexpected do: "+aTerm);
			}
			actionLock.notified = false;
		}			

		synchronized(expectLock){
			expectLock.notified = true;
			expectLock.notify();
		}
	}
	
	public ATerm doEval(ATerm aTerm){
		ATerm value;
		
		synchronized(actionLock){
			if(expecting){
				value = expectedEvals.get(aTerm);
				
				if(value == null) testCase.testFailed("Tool "+getToolName()+": unexpected eval: "+aTerm);
			}else{
				actionQueue.add(aTerm);
				
				long waitToMillis = System.currentTimeMillis() + DEFAULTTIMEOUT;
				while(!actionLock.notified && System.currentTimeMillis() < waitToMillis){
					try{
						long timeToWait = waitToMillis - System.currentTimeMillis();
						if(timeToWait > 0) actionLock.wait(timeToWait);
					}catch(InterruptedException irex){/* Ignore this.*/}
				}
				
				value = expectedEvals.get(aTerm);	
				
				if(value == null) testCase.testFailed("Tool "+getToolName()+": unexpected eval: "+aTerm);
			}
			
			actionLock.notified = false;
		}
		
		synchronized(expectLock){
			expectLock.notified = true;
			expectLock.notify();				
		}
		
		return value;
	}
	
	public ATerm doGetPerformanceStats(){
		return getPerformanceStats();
	}
	
	public void doReceiveAckEvent(ATerm aTerm){
		tool.receiveAckEvent(aTerm);
	}
	
	public void doTerminate(ATerm aTerm){
		tool.receiveTerminate(aTerm);
	}
	
	public void sendEvent(final ATerm event, final long timeout){
		final WaitLock eventLock = new WaitLock();
		final String eventName = ((ATermAppl) event).getAFun().toString();
		
		if(verbose) System.err.println("Tool "+getToolName()+" sends event: "+event);

		expectedAckEvents.put(eventName, eventLock);
		
		postEvent(event);
		
		new Thread(){
			public void run(){
				long waitToMillis = System.currentTimeMillis() + timeout;
			
				synchronized(eventLock){
					while(!eventLock.notified){
						try{
							long timeToWait = waitToMillis - System.currentTimeMillis();
							if(timeToWait > 0) eventLock.wait(timeToWait);
						}catch(InterruptedException irex){/* Ignore this. */}
						
						if(System.currentTimeMillis() >= waitToMillis){
							expectedAckEvents.remove(eventName);
							synchronized(completionLock){
								completionLock.notifyAll();
							}
							
							testCase.testFailed("Tool "+getToolName()+": timed out while waiting for ack event of: "+event);
						}
					}
				}
			}
		}.start();
	}
	
	public ATerm sendRequest(final ATerm request, ATerm expectedResponse, long timeout){
		final Object requestLock = new Object();
		
		if(verbose) System.err.println("Tool "+getToolName()+" sends request "+request);
		
		class RequestPoster extends Thread{
			public volatile ATerm response = null;
			
			public void run(){
				response = postRequest(request);
				synchronized(requestLock){
					requestLock.notify();
				}
			}
		}
		RequestPoster requestPoster = new RequestPoster();
		requestPoster.start();

		long waitToMillis = System.currentTimeMillis() + timeout;
		synchronized(requestLock){
			while(requestPoster.response == null){
				try{
					long timeToWait = waitToMillis - System.currentTimeMillis();
					if(timeToWait > 0) requestLock.wait(timeToWait);
				}catch(InterruptedException irex){/* Ignore this. */}
				
				if(System.currentTimeMillis() >= waitToMillis) testCase.testFailed("Tool "+getToolName()+": timed out while waiting for response of: "+request); 
			}
		}
		
		ATerm response = requestPoster.response;
		
		if(response != expectedResponse) testCase.testFailed("Tool "+getToolName()+": received an unexpected response. Expected: "+expectedResponse+", got: "+response);
		
		return response;
	}
	
	public void registerForDo(ATerm action){
		expectedDos.add(action);
	}
	
	public void registerForEval(ATerm action, ATerm expectedResult){
		expectedEvals.put(action, expectedResult);
	}
	
	public void expectAction(long timeout){
		if(expectedDos.size() + expectedEvals.size() == 0) testCase.testFailed("Can't expect nothing.");
		
		if(verbose) System.err.println("Tool "+getToolName()+" expects eval's: "+expectedEvals+" and do's: "+expectedDos);

		synchronized(actionLock){
			expecting = true;
			
			if(!actionQueue.isEmpty()){
				actionLock.notified = true;
				actionLock.notify();
			}
		}
		
		long waitToMillis = System.currentTimeMillis() + timeout;
		synchronized(expectLock){
			while(!expectLock.notified){
				try{
					long timeToWait = waitToMillis - System.currentTimeMillis();
					if(timeToWait > 0) expectLock.wait(timeToWait);
				}catch(InterruptedException irex){/* Ignore this. */}
				
				if(System.currentTimeMillis() >= waitToMillis) testCase.testFailed("Tool "+getToolName()+": timed out while waiting for evals and/or do's.");
			}
			expectLock.notified = false;
		}
		
		synchronized(actionLock){
			actionQueue.clear();
			expecting = false;
		}
		
		expectedDos.clear();
		expectedEvals.clear();
	}
	
	public void waitForCompletion(long timeout){
		long waitToMillis = System.currentTimeMillis() + timeout;
		
		synchronized(completionLock){
			while(!expectedAckEvents.isEmpty()){
				try{
					long timeToWait = waitToMillis - System.currentTimeMillis();
					if(timeToWait > 0) completionLock.wait(timeToWait);
				}catch(InterruptedException irex){/* Ignore this. */}
				
				if(System.currentTimeMillis() >= waitToMillis) testCase.testFailed("Tool "+getToolName()+": timed out while waiting for ack event(s) for: "+expectedAckEvents);
			}
		}
		
		if(!actionQueue.isEmpty()) testCase.testFailed("Tool "+getToolName()+": unexpected queued eval or do detected while waiting for completion."+actionQueue);
	}
	
	private static class WaitLock{
		public volatile boolean notified = false;
	}
	
	/**
	 * Gathers performance statistics about this tool, like memory usage and the user-/system-time
	 * spend per thread.
	 * 
	 * @return Performance statictics.
	 */
	private ATerm getPerformanceStats(){
		String type = getType();
		
		// Type stuff
		ATerm remote = termFactory.makeAppl(termFactory.makeAFun(type, 0, true));
		ATerm toolType = termFactory.makeAppl(termFactory.makeAFun("type", 1, false), remote);
		
		ATerm java = termFactory.makeAppl(termFactory.makeAFun("Java", 0, true));
		ATerm toolLanguage = termFactory.makeAppl(termFactory.makeAFun("language", 1, false), java);
		
		ATerm toolData = termFactory.makeAppl(termFactory.makeAFun("tool", 2, false), toolType, toolLanguage);
		
		// Memory stuff
		MemoryMXBean mmxb = ManagementFactory.getMemoryMXBean();
		long heapMemoryUsage = mmxb.getHeapMemoryUsage().getUsed();
		long nonHeapMemoryUsage = mmxb.getNonHeapMemoryUsage().getUsed();
		
		ATerm heapUsage = termFactory.makeAppl(termFactory.makeAFun("heap-usage", 1, false), termFactory.makeInt(((int) (heapMemoryUsage / 1024))));
		ATerm nonHeapUsage = termFactory.makeAppl(termFactory.makeAFun("non-heap-usage", 1, false), termFactory.makeInt(((int) (nonHeapMemoryUsage / 1024))));
		
		ATerm memory = termFactory.makeAppl(termFactory.makeAFun("memory-usage", 2, false), heapUsage, nonHeapUsage);
		
		// Thread stuff
		ThreadGroup currentThreadGroup = Thread.currentThread().getThreadGroup();
		Thread[] relevantThreads = new Thread[currentThreadGroup.activeCount() * 2 + 10]; // Create an array that's more then big enough.
		currentThreadGroup.enumerate(relevantThreads);
		
		Set<Long> relevantThreadIds = new HashSet<Long>();
		for(int i = 0; i < relevantThreads.length; i++){
			Thread thread = relevantThreads[i];
			if(thread != null){
				relevantThreadIds.add(new Long(thread.getId()));
			}
		}
		
		ThreadMXBean tmxb = ManagementFactory.getThreadMXBean();
		
		ATerm threads;
		
		long[] threadIds = tmxb.getAllThreadIds();
		int nrOfThreads = threadIds.length;
		try{
			ATermList threadsList = termFactory.makeList();
			for(int i = 0; i < nrOfThreads; i++){
				long threadId = threadIds[i];
				if(relevantThreadIds.contains(new Long(threadId))){ // Only list the info if we're interested in it.
					ThreadInfo ti = tmxb.getThreadInfo(threadId);
					if(ti != null){
						String threadName = ti.getThreadName();
						long userTime = tmxb.getThreadUserTime(threadIds[i]);
						long systemTime = tmxb.getThreadCpuTime(threadIds[i]) - userTime;
						
						if((userTime + systemTime) <= 0) continue;
						
						ATerm userTimeTerm = termFactory.makeAppl(termFactory.makeAFun("user-time", 1, false), termFactory.makeInt(((int) (userTime / 1000000))));
						ATerm systemTimeTerm = termFactory.makeAppl(termFactory.makeAFun("system-time", 1, false), termFactory.makeInt(((int) (systemTime / 1000000))));
						ATerm thread = termFactory.makeAppl(termFactory.makeAFun(threadName, 2, false), userTimeTerm, systemTimeTerm);
						
						threadsList = termFactory.makeList(thread, threadsList);
					}
				}
			}
			
			threads = termFactory.makeAppl(termFactory.makeAFun("threads", 1, false), threadsList);
		}catch(UnsupportedOperationException uoex){
			threads = termFactory.make("threads(unsupported-operation)");
			LoggerFactory.log("Thread time profiling is not supported by this JVM.", ILogger.ERROR, IToolBusLoggerConstants.TOOL);
		}
		
		return termFactory.makeAppl(termFactory.makeAFun("performance-stats", 3, false), toolData, memory, threads);
	}
}
